const ArtefactServer = require('artefact-server')
const { artefactEncryptionKey } = require('artefact-store')
const path = require('path')
const Obz = require('obz')
const defer = require('pull-defer')
const toPull = require('stream-to-pull-stream')
const pull = require('pull-stream')
const paraMap = require('pull-paramap')
const sort = require('pull-sort')
const { v4: uuid } = require('uuid') // TODO check which version to use!

const IsPrunable = require('./lib/is-prunable-file')

module.exports = {
  name: 'hyperBlobs',
  version: require('./package.json').version,
  manifest: {
    add: 'sink',
    driveAddress: 'async',
    registerDrive: 'async',
    prune: 'async'
  },
  permissions: {
    anonymous: {
      allow: ['registerDrive']
    }
  },
  init (ssb, config) {
    const {
      port = 26836,
      pataka = false,
      storeOpts = {}
    } = (config.hyperBlobs || {})

    const server = Obz() // artefact-server
    ArtefactServer({
      storage: path.resolve(config.path, 'hyperBlobs'),
      port,
      pataka,
      storeOpts
    })
      .then(_server => server.set(_server))
      .catch(err => { throw err })

    ssb.close.hook(function (close, args) {
      const _this = this
      server.once(server => {
        server.close()
          .then(() => {
            close.apply(_this, args)
          })
      })
    })

    if (!pataka) {
      // TODO have pataka register all the drive addresses it knows with peers who connect!

      const peerAttempts = {
        // peerId: attemptCount
      }
      ssb.on('rpc:connect', remote => {
        if (remote.id === ssb.id) return // could be ssb-client connection
        if (!remote.hyperBlobs) return

        driveAddress((err, address) => {
          if (err) throw err

          peerAttempts[remote.id] = peerAttempts[remote.id] || 0
          if (remote.closed) return reconnect(remote)
          // NOTE: the artefact-store takes a while to get ready
          // the rpc connecting can closes if there's is no action, here we manually start the connection again

          attemptRegister(remote, address)
        })
      })

      function reconnect (remote) {
        if (!ssb.conn) {
          console.error('ssb-hyper-blobs needs ssb-conn to restart closed connections')
          return
        }
        ssb.conn.connect(remote.stream.address)
      }
      function attemptRegister (remote, address) {
        peerAttempts[remote.id] = peerAttempts[remote.id] + 1

        remote.hyperBlobs.registerDrive(address, (err) => {
          if (!err) {
            delete peerAttempts[remote.id]
            return
          }
          if (peerAttempts[remote.id] < 5) {
            console.log(`remote.hyperBlobs.registerDrive failed (attempt ${peerAttempts[remote.id]})`)
            if (remote.closed) return reconnect(remote)
            return setTimeout(() => attemptRegister(remote, address), 2000)
          }

          console.error('hyperBlobs.registerDrive failed. Giving up', err)
        })
      }
    }

    // let _driveAddress
    function driveAddress (cb) {
      if (pataka) return cb(new Error('pataka does not have its own driveAddress'))

      server.once(server => {
        cb(null, toString(server.store.driveAddress))
      })
    }

    return {
      add (cb) {
        const sink = defer.sink()

        const blobId = uuid()
        const readKey = artefactEncryptionKey()
        const opts = {}

        server.once(server => {
          const writeStream = server.store.createWriteStream(blobId, readKey, opts)

          const _sink = toPull.sink(writeStream, (err) => {
            if (err) return cb(err)

            // TODO - check this is needed?
            server.store.fileReady(blobId)
              .then(() => {
                cb(null, {
                  driveAddress: toString(server.store.driveAddress),
                  blobId,
                  readKey: toString(readKey)
                  // fileName....
                })
              })
          })
          sink.resolve(_sink)
        })

        return sink
      },
      driveAddress,
      registerDrive (address, cb) {
        server.once(server => {
          server.store.addDrive(address)
            .then(() => cb(null, true))
            .catch(err => cb(err))
        })
      },
      prune (opts = {}, cb) {
        /*
          opts = {
            pruneSize: Number,
            minSize: Number,
            maxSize: Number,
            minDate: Date|Number,
            maxDate: Date|Number,
            sort: Function
          }
        */

        server.once(server => {
          const pruneFile = PruneFile(server)
          const isPrunable = IsPrunable(opts)
          // NOTE we must instantiate isPrunable here because it closes over a counter

          pull(
            pullFiles(server),
            sort(opts.sort || sortByLastAccessedDesc),
            pull.filter(isPrunable),
            paraMap(pruneFile, 6),
            pull.collect((err, prunedFiles) => {
              if (err) return cb(err)
              cb(null, prunedFiles)
            })
          )
        })
      }
    }
  }
}

function toString (buf) {
  if (typeof buf === 'string') return buf
  return buf.toString('hex')
}

function sortByLastAccessedDesc (fileA, fileB) {
  return fileA.atime - fileB.atime
}

function pullFiles (server) {
  const getFileStats = GetFileStats(server)

  return pull(
    pull.values(server.store.remoteDrives),
    pull.map(drive => drive.key),
    paraMap((driveAddress, cb) => {
      server.store.getFilesByDrive('/', driveAddress)
        .catch(err => cb(err))
        .then(files => cb(null, { driveAddress, files })) // here we save the driveAddress as well, its needed for later
    }, 6),
    paraMap(getFileStats, 6),
    pull.flatten()
  )
}

function GetFileStats (server) {
  return function ({ driveAddress, files }, cb) {
    pull(
      pull.values(files),
      paraMap((filename, cb) => {
        server.store.fileStats(filename, driveAddress)
          .catch(err => cb(err))
          .then(stats => cb(null, { filename, driveAddress, ...stats }))
      }, 6),
      pull.collect((err, files) => {
        if (err) return cb(err)
        cb(null, files)
      })
    )
  }
}

function PruneFile (server) {
  return function pruneFiles (file, cb) {
    const { driveAddress, filename } = file

    server.store.remove(filename, driveAddress)
      .catch(err => cb(err))
      .then(() => cb(null, file))
  }
}
